from frasco.models import db
from sqlalchemy.ext.declarative import declared_attr
from .data import is_eu_country, should_charge_vat, get_vat_rate


class EUVATModelMixin(object):
    @declared_attr
    def _eu_vat_country(cls):
        return db.deferred(db.Column('eu_vat_country', db.String), group='eu_vat')

    @declared_attr
    def eu_vat_number(cls):
        return db.deferred(db.Column(db.String), group='eu_vat')

    @declared_attr
    def eu_vat_rate(cls):
        return db.deferred(db.Column(db.Float), group='eu_vat')

    def should_charge_eu_vat(self):
        return should_charge_vat(self.eu_vat_country, self.eu_vat_number)

    @property
    def eu_vat_country(self):
        return self._eu_vat_country

    @eu_vat_country.setter
    def eu_vat_country(self, value):
        if is_eu_country(value):
            self._eu_vat_country = value.upper()
            self.eu_vat_rate = get_vat_rate(self.eu_vat_country)
        else:
            self._eu_vat_country = None
            self.eu_vat_rate = None


class EUVATInvoiceModelMixin(object):
    is_eu_country = db.Column(db.Boolean, default=False)
    eu_vat_number = db.Column(db.String)
    eu_exchange_rate = db.Column(db.Float)
    eu_vat_amount = db.Column(db.Float, default=0)



from flask import current_app
from frasco.ext import get_extension_state
from frasco.mail import send_mail
import pyotp

from .password import hash_password, generate_random_password


def generate_otp_code(user):
    user.otp_code = pyotp.random_base32()
    return generate_otpauth_url(user)


def enable_2fa(user, code):
    if not verify_2fa(user, code):
        return False
    user.two_factor_auth_enabled = True
    recovery_code = generate_otp_recovery_code(user)
    send_mail(user.email, 'users/2fa_enabled', recovery_code=recovery_code)
    return recovery_code


def disable_2fa(user):
    if not user.two_factor_auth_enabled:
        return
    user.two_factor_auth_enabled = False
    user.otp_code = None
    user.otp_recovery_code = None
    send_mail(user.email, 'users/2fa_disabled')


def verify_2fa(user, code):
    if not user.otp_code:
        return False
    totp = pyotp.TOTP(user.otp_code)
    return totp.verify(code)


def generate_otp_recovery_code(user):
    recovery_code = generate_random_password()
    user.otp_recovery_code = hash_password(recovery_code)
    return recovery_code


def generate_otpauth_url(user):
    return pyotp.totp.TOTP(user.otp_code).provisioning_uri(user.email,
        issuer_name=get_extension_state('frasco_users').options['2fa_issuer_name'] or current_app.config.get('TITLE'))



import click
import random


@click.command('gen-secret-key')
def gen_secret_key_command():
    """Generate a secret key for Flask session"""
    click.echo(''.join([random.SystemRandom().choice('abcdefghijklmnopqrstuvwxyz0123456789!@#$%^&*(-_=+)') for i in range(50)]))




class ApiError(Exception):
    def __init__(self, message, http_code=500):
        self.message = message
        self.http_code = http_code

    def __str__(self):
        return str(self.message)


class ApiInputError(ApiError):
    def __init__(self, message="Invalid Request"):
        super(ApiInputError, self).__init__(message, 400)


class ApiAuthentificationRequiredError(ApiError):
    def __init__(self, message="Authentification Required"):
        super(ApiAuthentificationRequiredError, self).__init__(message, 401)


class ApiNotAuthorizedError(ApiError):
    def __init__(self, message="Not Authorized"):
        super(ApiNotAuthorizedError, self).__init__(message, 403)


class ApiNotFoundError(ApiError):
    def __init__(self, message="Not Found"):
        super(ApiNotFoundError, self).__init__(message, 404)




from flask import session, g, current_app
from frasco.ext import *
from frasco.utils import shell_exec
from frasco.helpers import get_remote_addr
import geoip2.database
import os
import logging
import click
import sys


DOWNLOAD_URL = "https://download.maxmind.com/app/geoip_download"
COUNTRY_EDITION = "GeoLite2-Country"
CITY_EDITION = "GeoLite2-City"


class FrascoGeoip(Extension):
    name = "frasco_geoip"
    defaults = {"use_city_db": True,
                "db": "GeoLite2-City.mmdb",
                "license_key": None}

    def _init_app(self, app, state):
        app.add_template_global(geolocate_country)
        app.add_template_global(geolocate_city)

        if os.path.exists(state.options['db']):
            state.reader = geoip2.database.Reader(state.options['db'])
        else:
            logging.getLogger('frasco.geoip').info('%s does not exists, geolocation is disabled' % state.options['db'])

        @app.cli.command('dl-geo-db')
        @click.option('--edition')
        @click.option('--suffix', default='tar.gz')
        @click.option('--license-key')
        def download_db(edition=None, suffix='tar.gz', license_key=None):
            """Download GeoLite2 databases from MaxMind"""
            if not license_key:
                license_key = state.options['license_key']
            if not license_key:
                click.echo("You must register on MaxMind to obtain a license key in order to download databases")
                sys.exit(1)
            tmpfilename = "/tmp/GeoLite2.tar.gz"
            if not edition:
                edition = CITY_EDITION if state.options["use_city_db"] else COUNTRY_EDITION
            url = "%s?edition_id=%s&suffix=%s&license_key=%s" % (DOWNLOAD_URL, edition, suffix, license_key)
            shell_exec(["wget", "-O", tmpfilename, url])
            dbfilename = shell_exec(["tar", "tzf", tmpfilename, "--no-anchored", state.options['db']], echo=False).decode().strip()
            shell_exec(["tar", "-C", "/tmp", "-xzf", tmpfilename, dbfilename])
            shell_exec(["mv", os.path.join("/tmp", dbfilename), state.options['db']])


def geolocate(addr=None, method=None):
    state = get_extension_state('frasco_geoip')
    if not hasattr(state, 'reader'):
        return
    if not method:
        method = 'city' if state.options['use_city_db'] else 'country'
    try:
        return getattr(state.reader, method)(addr or get_remote_addr())
    except:
        pass


def geolocate_country(addr=None, use_session_cache=True):
    if use_session_cache and "geo_country_code" in session:
        return session["geo_country_code"]

    if get_extension_state('frasco_geoip').options['use_city_db']:
        city, country = geolocate_city(addr, use_session_cache)
        return country

    r = geolocate(addr, 'country')
    if r:
        if use_session_cache:
            session["geo_country_code"] = r.country.iso_code
        return r.country.iso_code


def geolocate_city(addr=None, use_session_cache=True):
    if use_session_cache and "geo_city" in session:
        return session["geo_city"]
    r = geolocate(addr, 'city')
    if r:
        if use_session_cache:
            session["geo_city"] = r.city.name
            session["geo_country_code"] = r.country.iso_code
        return r.city.name, r.country.iso_code
    return None, None


def clear_geo_cache():
    session.pop("geo_country_code", None)
    session.pop("geo_city", None)




from flask import json
from flask.config import Config as FlaskConfig
from frasco.utils import deep_update_dict
import os
import yaml
import errno
import logging


logger = logging.getLogger('frasco')


class Config(FlaskConfig):
    """Subclass of Flask's Config class to add support to load from YAML file
    """
    def from_mapping(self, *mapping, **kwargs):
        mappings = []
        if len(mapping) == 1:
            if hasattr(mapping[0], 'items'):
                mappings.append(list(mapping[0].items()))
            else:
                mappings.append(mapping[0])
        elif len(mapping) > 1:
            raise TypeError(
                'expected at most 1 positional argument, got %d' % len(mapping)
            )
        deep_update = kwargs.pop('_deep_update', False)
        mappings.append(list(kwargs.items()))
        for mapping in mappings:
            if deep_update:
                deep_update_dict(self, dict((k.upper(), v) for (k, v) in mapping))
            else:
                for (key, value) in mapping:
                    self[key.upper()] = value
        return True

    def from_file(self, filename, load=None, silent=False, deep_update=False):
        if not load:
            if filename.endswith(".py"):
                return self.from_pyfile(filename, silent)
            if filename.endswith(".js") or filename.endswith(".json"):
                load = json.load
            if filename.endswith(".yml") or filename.endswith(".yaml"):
                load = yaml.safe_load

        filename = os.path.join(self.root_path, filename)

        try:
            with open(filename) as f:
                obj = load(f)
        except OSError as e:
            if silent and e.errno in (errno.ENOENT, errno.EISDIR):
                return False

            e.strerror = f"Unable to load configuration file ({e.strerror})"
            raise

        return self.from_mapping(obj, _deep_update=deep_update)

    def from_json(self, filename, **kwargs):
        return self.from_file(filename, json.load, **kwargs)

    def from_yaml(self, filename, **kwargs):
        return self.from_file(filename, yaml.safe_load, **kwargs)


def load_config(config, config_filename='config.yml', env=None, deep_update=False, logger=None):
    if logger is None:
        logger = logging.getLogger('frasco')
    if os.path.exists(config_filename):
        if logger:
            logger.info('Loading config from %s' % config_filename)
        config.from_file(config_filename, deep_update=deep_update)
    if not env:
        return
    filename, ext = os.path.splitext(config_filename)
    env_filename = filename + "-" + env + ext
    if os.path.exists(env_filename):
        if logger:
            logger.info('Loading config from %s' % env_filename)
        config.from_file(env_filename, deep_update=True)


def load_app_config(app, **kwargs):
    kwargs.setdefault('env', app.config['ENV'])
    load_config(app.config, **kwargs)


def update_config_with_env_vars(config, prefix, logger=None):
    if logger is None:
        logger = logging.getLogger('frasco')
    config = {}
    prefix = prefix.upper()
    for k, v in os.environ.items():
        if k.startswith(prefix + "_"):
            config[k[len(prefix)+1:]] = v
    if config:
        if logger:
            logger.info('Using config from environment variables')
        config.update(config)


def update_app_config_with_env_vars(app, prefix):
    update_config_with_env_vars(app.config, prefix)
